from durabledict.encoding import DefaultEncoding, EncodingError, DecodingError


class DurableDict(object):
    """
    Dictionary that calls out to its persistant data store when items are
    created or deleted.  Syncs with data fron the data store before every read,
    unless ``autosync=False`` is passed, which causes the dict to only sync
    data from the data store on writes and when ``sync()`` is called.

    By default, objects are encoded in the durable store using
    ``encoding.PickleEncoding``, but that can be changed by passing in another
    encoder in as the ``encoding`` kwarg.

    If you need to switch between two encodings provide the old_encoding as a fallback.
    """

    def __init__(self, autosync=True, encoding=DefaultEncoding, old_encoding=None):
        self.__dict = dict()
        self.last_synced = 0
        self.autosync = autosync
        self.encoding = encoding
        self.old_encoding = old_encoding
        self.__sync_with_durable_storage(force=True)

    @property
    def cache_expired(self):
        persistance_last_updated = self.last_updated()

        if not self.last_synced or persistance_last_updated > self.last_synced:
            return persistance_last_updated

    def sync(self):
        self.__sync_with_durable_storage(force=True)

    def pop(self, key, default=None):
        result = self._pop(key, default)
        self.__sync_with_durable_storage(force=True)
        return result

    def setdefault(self, key, default=None):
        result = self._setdefault(key, default)
        self.__sync_with_durable_storage(force=True)
        return result

    def get(self, key, default=None):
        self.__sync_with_durable_storage()
        return self.__dict.get(key, default)

    def __setitem__(self, key, val):
        self.persist(key, val)
        self.__sync_with_durable_storage(force=True)

    def __delitem__(self, key):
        self.depersist(key)
        self.__sync_with_durable_storage(force=True)

    def __getitem__(self, key):
        self.__sync_with_durable_storage()
        return self.__dict.__getitem__(key)

    def __getattr__(self, name):
        self.__sync_with_durable_storage()
        return getattr(self.__dict, name)

    def __len__(self):
        self.__sync_with_durable_storage()
        return self.__dict.__len__()

    def __cmp__(self, other):
        self.__sync_with_durable_storage()
        return self.__dict.__cmp__(other)

    def __repr__(self):
        return self.__dict.__repr__()

    def __contains__(self, key):
        self.__sync_with_durable_storage()
        return self.__dict.__contains__(key)

    def __sync_with_durable_storage(self, force=False):
        if not self.autosync and not force:
            return

        cache_expired_at = self.cache_expired

        if cache_expired_at:
            self.__dict = self.durables()
            self.last_synced = cache_expired_at

    def _encode(self, val):
        try:
            return self.encoding.encode(val)
        except EncodingError:
            if self.old_encoding:
                return self.old_encoding.encode(val)
            else:
                raise

    def _decode(self, val):
        try:
            return self.encoding.encode(val)
        except EncodingError:
            if self.old_encoding:
                return self.old_encoding.encode(val)
            else:
                raise


    def persist(self, key, val):
        raise NotImplementedError

    def depersist(self, key):
        raise NotImplementedError

    def durables(self):
        raise NotImplementedError

    def last_updated(self):
        raise NotImplementedError


class ConnectionDurableDict(DurableDict):
    """Base for Durable Dict classes that are connection oriented."""

    def __init__(self, keyspace=None, connection=None, **kwargs):
        self.keyspace = keyspace
        self.connection = connection

        self.connection_hook()

        super(ConnectionDurableDict, self).__init__(**kwargs)

    def connection_hook(self):
        pass



from base import DurableDict


class ModelDict(DurableDict):

    """
    Dictionary-style access to a model. Populates a cache and a local in-memory
    to avoid multiple hits to the database.

        # Given ``Model`` that has a primary key (``pk``) column that can be
        # strings:

        mydict = ModelDict(Model.manager, value_col='foo')
        mydict['test']
        >>> 'bar' #doctest: +SKIP

    The first positional argument to ``ModelDict`` is ``manager``, which is an
    instance of a Manager which ``ModelDict`` uses to read and write to your
    database.  Any object that conforms to the interface can work, but the
    expectation is that ``manager`` is a Django.model manager.

    The constructor also takes a 2nd positional ``cache`` argument, which is an
    object that responds to two methods, add and incr.  The cache object is used
    to manage the value for last_updated.  ``add`` is called on initialize to
    create the key if it does not exist with the default value, and ``incr`` is
    done to atomically update the last_updated value.

    By default, ``ModelDict`` will use the ``pk`` column of your model as the
    "key" for the dictionary.  If you want to use another key besides the
    ``Model``s ``pk``, you may specify that in the constructor with ``key_col``
    kwarg.  For instance, if your ``Model`` has a column called ``id``, you can
    index into that column by passing ``key_col='id'`` in to the contructor:

        mydict = ModelDict(Model, key_col='id', value_col='foo')
        mydict['test']
        >>> 'bar' #doctest: +SKIP

    Additionally, the default column used to store the values is the ``value``
    column.  If you'd like to use another column pass the ``value_col`` kwarg to
    the constructor.

    By default, ``ModelDict`` instances will return the decoded value present in
    the ``value_col``.  If, instead, you would like to return the entire model
    instance, you can instead pass ``True`` to the ``return_instances`` kwarg.
    """

    #: The value to increate last_updated by, in the event that the key is
    # missing from the cache.  See the comment inside touch_last_updated for
    # details.
    LAST_UPDATED_MISSING_INCREMENT = 1000

    def __init__(
        self,
        manager=None,
        cache=None,
        key_col='key',
        value_col='value',
        return_instances=False,
        **kwargs
    ):
        self.manager = manager
        self.cache = cache
        self.cache_key = 'last_updated'
        self.return_instances = return_instances
        self.key_col = key_col
        self.value_col = value_col

        self.cache.add(self.cache_key, 1)  # Only adds if key does not exist

        super(ModelDict, self).__init__(**kwargs)

    def persist(self, key, val):
        instance, created = self.get_or_create(key, val)

        if not created and getattr(instance, self.value_col) != val:
            setattr(instance, self.value_col, self.encoding.encode(val))
            instance.save()

        self.touch_last_updated()

    def depersist(self, key):
        self.manager.get(**{self.key_col: key}).delete()
        self.touch_last_updated()

    def durables(self):
        if self.return_instances:
            return dict((i.key, i) for i in self.manager.all())
        else:
            encoded_tuples = self.manager.values_list(
                self.key_col,
                self.value_col
            )
            return dict((k, self.encoding.decode(v)) for k, v in encoded_tuples)

    def _setdefault(self, key, default=None):
        instance, created = self.get_or_create(key, default)

        if created:
            self.touch_last_updated()

        return self.encoding.decode(getattr(instance, self.value_col))

    def _pop(self, key, default=None):
        try:
            instance = self.manager.get(**{self.key_col: key})
            value = self.encoding.decode(getattr(instance, self.value_col))
            instance.delete()
            self.touch_last_updated()
            return value
        except self.manager.model.DoesNotExist:
            if default is not None:
                return default
            else:
                raise KeyError

    def get_or_create(self, key, val):
        return self.manager.get_or_create(
            defaults={self.value_col: self.encoding.encode(val)},
            **{self.key_col: key}
        )

    def last_updated(self):
        return self.cache.get(self.cache_key)

    def touch_last_updated(self):
        try:
            self.cache.incr(self.cache_key)
        except ValueError:
            # The last_updated cache key is missing. This may be because it has
            # expired or been explicitly deleted. It is then necessary to
            # recreate the value. by synthesizing an ``incr``. This is
            # accomplished by adding the current ``self.last_synced`` value plus
            # some increment, naively "1".
            #
            # However, there is a race condition. It is entirely possible that
            # an instance of DurableDict which sees the key as missing may
            # have an out of date value for ``self.last_synced`` - such as if
            # multiple updates have occurred from other instances of DurableDict
            # (and are reflected in the cache value of last_udated), but this
            # instance of DurableDict has not seen those changes yet. If the
            # cache key is deleted before this instance hasn't seen those
            # changed, but this instance sees the cache as expired, then it will
            # update the value to what it knows to be last_synced + 1, which is
            # incorrect.
            #
            # A workaround to this problem is to set the new cache value to what
            # this instance thinks last_synced is + some non-trivial number.
            # This is not a perfect solution, as it's possible there could be
            # LAST_UPDATED_MISSING_INCREMENT changes to the dict since this
            # instance last checked, so LAST_UPDATED_MISSING_INCREMENT should
            # be set to a value high enough to where that possibility is
            # unlikely.
            added_key = self.cache.add(
                self.cache_key,
                self.last_synced + self.LAST_UPDATED_MISSING_INCREMENT
            )

            # XXX There is still  race condition here. It is possible that the
            # key can be deleted between the call to ``add`` and ``incr``, in
            # which case the ``incr`` call will fail.  This is a unlikely
            # scenario and as such is not handled.

            # If the ``add`` did not succeed, it means that another instance of
            # DurableDict beat this one to adding the cache value back and so
            # this instance should do the ``incr`` like normal.
            if not added_key:
                self.cache.incr(self.cache_key)



from base import ConnectionDurableDict


class RedisDict(ConnectionDurableDict):

    """
    Dictionary-style access to a redis hash table. Populates a cache and a local
    in-memory to avoid multiple hits to the database.

    Functions just like you'd expect it::

        mydict = RedisDict('my_redis_key', Redis())
        mydict['test']
        >>> 'bar' #doctest: +SKIP
    """

    def __init__(self, **kwargs):
        super(RedisDict, self).__init__(**kwargs)
        self.__touch_last_updated()

    def persist(self, key, value):
        encoded = self.encoding.encode(value)
        self.__touch_and_multi(('hset', (self.keyspace, key, encoded)))

    def depersist(self, key):
        self.__touch_and_multi(('hdel', (self.keyspace, key)))

    def durables(self):
        encoded = self.connection.hgetall(self.keyspace)
        tuples = [(k, self.encoding.decode(v)) for k, v in encoded.items()]
        return dict(tuples)

    def last_updated(self):
        return int(self.connection.get(self.__last_update_key) or 0)

    # TODO: setdefault always touches the last_updated value, even if the key
    # existed already.  It should only touch last_updated if the key did not
    # already exist
    def _setdefault(self, key, default=None):
        encoded = self.__touch_and_multi(
            ('hsetnx', (self.keyspace, key, self.encoding.encode(default))),
            ('hget', (self.keyspace, key)),
            returns=-1
        )
        return self.encoding.decode(encoded)

    def _pop(self, key, default=None):
        last_updated, encoded, key_existed = self.__touch_and_multi(
            ('hget', (self.keyspace, key)),
            ('hdel', (self.keyspace, key))
        )

        if key_existed:
            return self.encoding.decode(encoded)
        elif default:
            return default
        else:
            raise KeyError

    def __touch_and_multi(self, *args, **kwargs):
        """
        Runs each tuple tuple of (redis_cmd, args) in provided inside of a Redis
        MULTI block, plus an increment of the last_updated value, then executes
        the MULTI block.  If ``returns`` is specified, it returns that index
        from the results list.  If ``returns`` is None, returns all values.
        """

        with self.connection.pipeline() as pipe:
            pipe.incr(self.__last_update_key)
            [getattr(pipe, function)(*a) for function, a in args]
            results = pipe.execute()

            if kwargs.get('returns'):
                return results[kwargs.get('returns')]
            else:
                return results

    def __touch_last_updated(self):
        return self.connection.incr(self.__last_update_key)

    @property
    def __last_update_key(self):
        return self.keyspace + 'last_updated'



from durabledict.base import DurableDict
from durabledict.encoding import NoOpEncoding


class MemoryDict(DurableDict):

    '''
    Does not actually persist any data to a persistant storage.  Instead, keeps
    everything in memory.  This is really only useful for use in tests
    '''

    def __init__(self, **kwargs):
        self.__storage = dict()
        self.__last_updated = 1
        if 'encoding' not in kwargs:
            kwargs['encoding'] = NoOpEncoding

        super(MemoryDict, self).__init__(**kwargs)

    def persist(self, key, val):
        self.__storage[key] = self.encoding.encode(val)
        self.__last_updated += 1

    def depersist(self, key):
        del self.__storage[key]
        self.__last_updated += 1

    def durables(self):
        encoded_tuples = self.__storage.items()
        tuples = [(k, self.encoding.decode(v)) for k, v in encoded_tuples]
        return dict(tuples)

    def last_updated(self):
        return self.__last_updated

    def _setdefault(self, key, default=None):
        self.__last_updated += 1
        val = self.__storage.setdefault(key, self.encoding.encode(default))
        return self.encoding.decode(val)

    def _pop(self, key, default=None):
        self.__last_updated += 1

        if default:
            default = self.encoding.encode(default)

        val = self.__storage.pop(key, default)

        if val is None:
            raise KeyError

        return self.encoding.decode(val)



from durabledict.base import ConnectionDurableDict

from functools import wraps
import posixpath


def validate_key(func):
    """
    Decorator to validate a key for zookeeper.
    """

    @wraps(func)
    def wrapper(self, key, *args, **kwargs):
        if posixpath.sep in key:
            raise ValueError('Keys cannot contains slashes')

        return func(self, key, *args, **kwargs)

    return wrapper


class ZookeeperDict(ConnectionDurableDict):

    """
    Dictionary backed by Zookeeper.  Functions just as you would expect a normal
    dictiony to function, except the values in the dictionary are persisted and
    loaed from Zookeper located at the spcified ``path`` in the constructor.

    Due to Zookeeper's watchers, this dictionary has the nice property that the
    called to its ``last_updated`` method (to check if the storage has been
    updated since the dict was last synced) simply returns a cached value -- it
    does not query Zookeeper or anything like that, so it's basically free.
    This means that you can run this dictionary with ``autosync=True`` and it
    will still be reasonably performant.

    Dictionary keys in a ``ZookeeperDict`` are stored at inividual nodes in the
    zookeeper heirarchy, with the value of the node being the value of that key.
    Each node for each dict key is a child of the "root" node, whose path is
    specified with the ``path`` argument in the constructor.

        >>> from durabledict.dict import ZookeeperDict
        >>> from kazoo.client import KazooClient
        >>> kazoo = KazooClient()
        >>> kazoo.start()
        >>> zkdict = ZookeeperDict(kazoo, '/app/config')
        >>> zkdict['exchange_rate'] = 25
        >>> zkdict['language'] = 'en-US'
        >>> zkdict['exchange_rate']
        25
        >>> zkdict['language']
        'en-US'
        >>> zkdict.pop('exchange_rate')
        25
        >>> zkdict['exchange_rate']
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "durabledict/base.py", line 53, in __getitem__
            return self.__dict.__getitem__(key)
        KeyError: 'exchange_rate'

    NOTE: unlike ``RedisDict`` or ``Durabledict``, which are backed by hightly
    consistent backend storages, ``ZookeeperDict`` is backed with Zookeeper,
    which has looser consistency guarantees.

    Please see this page in the Zookeeper docs for details:

    http://zookeeper.apache.org/doc/r3.1.2/zookeeperProgrammers.html#ch_zkGuarantees

    The basic things to keep in mind are:

        Sequential Consistency:
        Updates from a client will be applied in the order that they were sent.

        Atomicity:
        Updates either succeed or fail -- there are no partial results.

        Single System Image:
        A client will see the same view of the service regardless of the server
        that it connects to.

        Reliability:
        Once an update has been applied, it will persist from that time forward
        until a client overwrites the update. This guarantee has two
        corollaries:

            If a client gets a successful return code, the update will have been
            applied. On some failures (communication errors, timeouts, etc) the
            client will not know if the update has applied or not. We take steps
            to minimize the failures, but the only guarantee is only present
            with successful return codes. (This is called the monotonicity
            condition in Paxos.)

            Any updates that are seen by the client, through a read request or
            successful update, will never be rolled back when recovering from
            server failures.

        Timeliness:
        The clients view of the system is guaranteed to be up-to-date within a
        certain time bound. (On the order of tens of seconds.) Either system
        changes will be seen by a client within this bound, or the client will
        detect a service outage.

    The cliff notes version of this is that a client's view of the world will
    always be consistent (you can read your own writes), but updates from other
    clients can take time to propogate to other clients.
    """

    def __init__(self, **kwargs):
        """
        Construct a new instance of a ``ZookeeperDict``.

        :param connection: Zookeeper client, likely ``KazooClient``
        :type connection: KazooClient
        :param keyspace: The path to the root config node.
        :type keyspace: string
        :param autosync: Sync with Zookeeper before each read.
        :type autosync: bool
        """
        super(ZookeeperDict, self).__init__(**kwargs)

    def connection_hook(self):
        if not self.connection.connected:
            self.connection.start()

        self.connection.retry(self.connection.ensure_path, self.keyspace)
        self._last_updated = None

        # TODO: The base DurableDict class updates last_updated itself
        # manually when adding a new key with __setattr__, as well as this watch
        # also incrementing the value.
        self.child_watch = self.connection.retry(
            self.connection.ChildrenWatch,
            self.keyspace,
            self.__increment_last_updated
        )

    @property
    def no_node_error(self):
        from kazoo.exceptions import NoNodeError
        return NoNodeError

    def last_updated(self):
        """
        Ever-increasing integer, which is bumped any time a key in Zookeeper has
        been changed (created, updated, deleted).

        The value in incremented manually by an instances when updating the
        dict, as well as when other instances of the dict update persistant
        storage, via a Zookeeper watch on the root config node.
        """
        return self._last_updated

    @validate_key
    def persist(self, key, value):
        """
        Encode and save ``value`` at ``key``.

        :param key: Key to store ``value`` at in Zookeeper.
        :type key: string
        :param value: Value to store. Encoded before being stored.
        :type value: value
        """
        encoded = self.encoding.encode(value)
        self.__set_or_create(key, encoded)
        self.__increment_last_updated()

    @validate_key
    def depersist(self, key):
        """
        Remove ``key`` from dictionary.

        :param key: Key to remove from Zookeeper.
        :type key: string
        """
        self.connection.retry(self.connection.delete, self.__path_of(key))
        self.__increment_last_updated()

    def durables(self):
        """
        Dictionary of all keys and their values in Zookeeper.
        """
        results = dict()

        for child in self.connection.retry(self.connection.get_children, self.keyspace):
            value, _ = self.connection.retry(
                self.connection.get,
                self.__path_of(child),
                watch=self.__increment_last_updated
            )
            results[child] = self.encoding.decode(value)

        return results

    @validate_key
    def _pop(self, key, default=None):
        """
        If ``key`` is present in Zookeeper, removes it from Zookeeper and
        returns the value.  If key is not in Zookeper and ``default`` argument
        is provided, ``default`` is returned.  If ``default`` argument is not
        provided, ``KeyError`` is raised.

        :param key: Key to remove from Zookeeper
        :type key: string
        :param default: Default object to return if ``key`` is not present.
        :type default: object
        """
        path = self.__path_of(key)
        value = None

        try:
            # We need to both delete and return the value that was in ZK here.
            raw_value, _ = self.connection.retry(self.connection.get, path)
            value = self.encoding.decode(raw_value)
        except self.no_node_error:
            # The node is already gone, so if a default is given, return it,
            # otherwise, raise KeyError
            if default:
                return default
            else:
                raise KeyError

        # Made it this far, it means have a value from the node and it existed
        # at least by that point in time
        try:
            # Try to delete the node
            self.connection.retry(self.connection.delete, path)
            self.__increment_last_updated()
        except self.no_node_error:
            # Someone deleted the node in the mean time...how nice!
            pass

        return value

    @validate_key
    def _setdefault(self, key, default=None):
        """
        If ``key`` is not present, set it as ``default`` and return it.  If
        ``key`` is present, return its value.

        :param key: Key to add to Zookeeper
        :type key: string
        :param default: Default object to return if ``key`` is present.
        :type default: object

        Will retry trying to get or create a node based on the "retry" config
        from the Kazoo client.
        """
        return self.connection.retry(self.__inner_set_default, key, default)

    def __path_of(self, key):
        return posixpath.join(self.keyspace, key)

    def __set_or_create(self, key, value):
        path = self.__path_of(key)
        self.connection.retry(self.connection.ensure_path, path)
        self.connection.retry(self.connection.set, path, value)

    def __increment_last_updated(self, children=None):
        if self._last_updated is None:
            self._last_updated = 0

        self._last_updated += 1

    def __inner_set_default(self, key, value):
        """
        Tries to return the value at key.  If the key does not exist, attempts
        to create it with the value.  If the node is created in the mean time,
        a ``NodeExistsError`` will be raised.
        """
        path = self.__path_of(key)

        try:
            # Try to get and return the existing node with its data
            value, _ = self.connection.retry(self.connection.get, path)
            return self.encoding.decode(value)
        except self.no_node_error:
            # Node does not exist, we have to create it
            self.connection.retry(self.connection.create, path, self.encoding.encode(value))
            self.__increment_last_updated()
            return value



import pickle
import base64
from functools import wraps


try:
    import simplejson as json
except ImportError:
    import json


class EncoderError(ValueError):
    pass


class EncodingError(EncoderError):
    pass


class DecodingError(EncoderError):
    pass


class Encoder(object):
    encoding_exceptions = ()
    decoding_exceptions = ()

    @property
    def encoder(self):
        raise NotImplementedError()

    @property
    def decoder(self):
        raise NotImplementedError()

    def encode(self, *args, **kwargs):
        try:
            return self.encoder(*args, **kwargs)
        except self.encoding_exceptions as e:
            raise EncodingError(e)

    def decode(self, *args, **kwargs):
        try:
            return self.decoder(*args, **kwargs)
        except self.decoding_exceptions as e:
            raise EncodingError(e)


class NoOpEncoding(Encoder):
    encode = staticmethod(lambda d: d)
    decode = staticmethod(lambda d: d)


class PickleEncoding(Encoder):
    encoding_exceptions = (pickle.PicklingError, )
    decoding_exceptions = (pickle.UnpicklingError, )

    @staticmethod
    def encoder(data):
        pickled = pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
        return base64.encodestring(pickled)


    @staticmethod
    def decoder(data):
        pickled = base64.decodestring(data)
        return pickle.loads(pickled)


class JSONEncoding(Encoder):
    encoding_exceptions = (ValueError, )
    decoding_exceptions = (ValueError, )

    encoder = json.dumps
    decoder = json.loads


DefaultEncoding = PickleEncoding



#!/usr/bin/env python

import os
import sys
import glob
from setuptools import setup


def get_zookeeper_paths():
    zk_path = None

    if sys.platform == 'darwin':
        homebrew_path = glob.glob('/usr/local/Cellar/zookeeper/*/libexec/')
        if homebrew_path:
            zk_path = homebrew_path[0]
    elif sys.platform == 'linux2':
        linux_default = '/usr/share/java/'
        if os.path.exists(linux_default):
            zk_path = linux_default

    if zk_path:
        return zk_path

    raise Exception('ZOOKEEPER_PATH must be in environment for tests to run')

if 'ZOOKEEPER_PATH' not in os.environ:
    os.environ['ZOOKEEPER_PATH'] = get_zookeeper_paths()

try:
    import multiprocessing  # Seems to fix http://bugs.python.org/issue15881
except ImportError:
    pass


setup(
    name='durabledict',
    version='0.9.0',
    author='DISQUS',
    author_email='opensource@disqus.com',
    url='http://github.com/disqus/durabledict/',
    description='Dictionary-style access to different types of models.',
    packages=['durabledict'],
    zip_safe=False,
    tests_require=[
        'Django',
        'nose',
        'mock',
        'redis',
        'kazoo',
    ],
    test_suite='nose.collector',
    include_package_data=True,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)
